/**
 * config - functions to read, write &co the uci config files
 *
 * Copyright (C) 2018-2019 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: Reidar Cederqvist <reidar.cederqvist@gmail.com>
 * Author: Ionut-Alex Oprea <ionutalexoprea@gmail.com>
 * Author: Jakob Olsson <jakob.olsson@iopsys.eu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 **/

#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>

#include <uci.h>
#include <libubox/blobmsg.h>
#include <libubox/list.h>
#include <json-c/json.h>

#include <linux/if.h>

#include "netmoded.h"
#include "config.h"

#define EQUAL (0)

static const char * const filtered_options[] = {
	".anonymous",
	".type",
	".name",
	".index",
	"device",
	"network",
	"ifname"
};

LIST_HEAD(ignored_options);

struct ignored_option {
	char *option;
	struct list_head list;
};

const char *json_get_string(struct json_object *object, const char *key)
{
	json_bool ret;
	struct json_object *value;

	if (!object || !json_object_is_type(object, json_type_object))
		return NULL;

	ret = json_object_object_get_ex(object, key, &value);
	if (!ret || !value || !json_object_is_type(value, json_type_string))
		return NULL;

	return json_object_get_string(value);
}

bool config_get_daisy_chain_enabled(void)
{
	char *errstr;
	struct uci_context *ctx;
	struct uci_element *e;
	struct uci_option *option;
	struct uci_section *section;
	struct uci_package *package;
	const char *daisy_chain;
	int rv;
	bool enabled = false;

	ctx = uci_alloc_context();
	if (!ctx)
		goto out;

	rv = uci_load(ctx, "wifilife", &package);
	if (rv != UCI_OK) {
		uci_get_errorstr(ctx, &errstr, NULL);
		ERROR("%s", errstr);
		free(errstr);
		goto out_ctx;
	}

	uci_foreach_element(&package->sections, e) {
		section = uci_to_section(e);
		if (strcmp(section->type, "cntlr"))
			continue;

		daisy_chain = uci_lookup_option_string(ctx, section, "daisy_chain");
		if (daisy_chain && !strncmp(daisy_chain, "1", 2))
			enabled = true;

		break;
	}

	rv = uci_unload(ctx, package);
	if (rv != UCI_OK) {
		uci_get_errorstr(ctx, &errstr, NULL);
		ERROR("%s", errstr);
		free(errstr);
	}
out_ctx:
	uci_free_context(ctx);
out:
	return enabled;
}

/* TODO: separate some functionality and use config_get_wifilife &
	config_get_wireless as wrappers */
int config_get_wifilife(struct blob_buf *buf)
{
	int rv = -1, section_index = 0;
	char *errstr;
	struct uci_context *ctx;
	struct uci_element *se, *oe, *le;
	struct uci_package *package;
	struct uci_section *section;
	struct uci_option *option;
	void *vt, *ot, *la;

	ctx = uci_alloc_context();
	if (!ctx)
		goto out;

	rv = uci_load(ctx, "wifilife", &package);
	if (rv != UCI_OK) {
		uci_get_errorstr(ctx, &errstr, NULL);
		ERROR("%s", errstr);
		free(errstr);
		goto out_ctx;
	}

	vt = blobmsg_open_table(buf, "values");

	/* for each section */
	uci_foreach_element(&package->sections, se) {

		section = uci_to_section(se);

		ot = blobmsg_open_table(buf, section->e.name);

		blobmsg_add_u8(buf, ".anonymous", section->anonymous);
		blobmsg_add_string(buf, ".type", section->type);
		blobmsg_add_string(buf, ".name", section->e.name);
		blobmsg_add_u32(buf, ".index", section_index++);

		/* for each option */
		uci_foreach_element(&section->options, oe) {
			option = uci_to_option(oe);
			switch (option->type) {
			case UCI_TYPE_STRING:
				blobmsg_add_string(buf, option->e.name,
						option->v.string);
				break;
			case UCI_TYPE_LIST:
				la = blobmsg_open_array(buf, option->e.name);
				/* for each value */
				uci_foreach_element(&option->v.list, le)
					blobmsg_add_string(buf, NULL, le->name);
				blobmsg_close_array(buf, la);
				break;
			default:
				break;
			}
		}

		blobmsg_close_table(buf, ot);
	}

	blobmsg_close_table(buf, vt);
	rv = uci_unload(ctx, package);
	if (rv != UCI_OK) {
		uci_get_errorstr(ctx, &errstr, NULL);
		ERROR("%s", errstr);
		free(errstr);
	}

out_ctx:
	uci_free_context(ctx);
out:
	return rv;
}

static bool is_ignored_option(const char *option)
{
	struct ignored_option *ign;

	list_for_each_entry(ign, &ignored_options, list) {
		if (strcmp(ign->option, option) == 0)
			return true;
	}

	return false;
}

/* parse uci wireless config into wifi ifaces */
int config_get_wireless(struct blob_buf *buf)
{
	int rv = -1, section_index = 0;
	char *errstr;
	struct uci_context *ctx;
	struct uci_element *se, *oe, *le;
	struct uci_package *package;
	struct uci_section *section;
	struct uci_option *option;
	void *vt, *ot, *la;
	char bssid[18];
	bool daisy_chain;

	ctx = uci_alloc_context();
	if (!ctx)
		goto out;

	rv = uci_load(ctx, "wireless", &package);
	if (rv != UCI_OK) {
		uci_get_errorstr(ctx, &errstr, NULL);
		ERROR("%s", errstr);
		free(errstr);
		goto out_ctx;
	}

	vt = blobmsg_open_table(buf, "values");

	/* for each section */
	uci_foreach_element(&package->sections, se) {
		section = uci_to_section(se);

		ot = blobmsg_open_table(buf, section->e.name);

		blobmsg_add_u8(buf, ".anonymous", section->anonymous);
		blobmsg_add_string(buf, ".type", section->type);
		blobmsg_add_string(buf, ".name", section->e.name);
		blobmsg_add_u32(buf, ".index", section_index++);

		/* for each option */
		uci_foreach_element(&section->options, oe) {
			option = uci_to_option(oe);
			if (is_ignored_option(option->e.name))
				continue;

			switch (option->type) {
			case UCI_TYPE_STRING:
				blobmsg_add_string(buf, option->e.name,
						option->v.string);
				break;
			case UCI_TYPE_LIST:
				la = blobmsg_open_array(buf, option->e.name);
				/* for each value */
				uci_foreach_element(&option->v.list, le)
					blobmsg_add_string(buf, NULL, le->name);
				blobmsg_close_array(buf, la);
				break;
			default:
				break;
			}
		}

		blobmsg_close_table(buf, ot);
	}

	blobmsg_close_table(buf, vt);

	rv = uci_unload(ctx, package);
	if (rv != UCI_OK) {
		uci_get_errorstr(ctx, &errstr, NULL);
		ERROR("%s", errstr);
		free(errstr);
	}

out_ctx:
	uci_free_context(ctx);
out:
	return rv;
}

/* get the band for the specified device (by string name)*/
static const char *config_wireless_get_band(struct json_object *values,
		const char *devicename)
{
	struct json_object *device;

	if (!json_object_object_get_ex(values, devicename, &device))
		return NULL;

	return json_get_string(device, "band");
}

/* get the device that runs on the specified band */
static const char *config_wireless_get_devicename(struct json_object *values,
		const char *band)
{
	const char *type, *b;

	json_object_object_foreach(values, key, val) {
		type = json_get_string(val, ".type");
		if (!type || strcmp(type, "wifi-device") != 0)
			continue;

		b = json_get_string(val, "band");
		if (b && strcmp(b, band) == 0)
			return key;
	}
	return NULL;
}

static int config_count_radio(struct json_object *values)
{
	int count = 0, rv;
	const char *type;

	if (!json_object_is_type(values, json_type_object))
		goto out;

	json_object_object_foreach(values, key, val) {
		type = json_get_string(val, ".type");
		if (type && strcmp(type, "wifi-device") == 0)
			++count;
	}

out:
	return count;
}

/* count the number of wireless interface
 * on band a or b or both (if band=="all")
 */
static void config_count_wifi_ifaces(struct json_object *values,
		const char *band, struct json_object *ifaces, const char *network)
{
	const char *type, *mode, *disabled, *device, *devicename = NULL;
	const char *band_string, *net;
	struct json_object *band_option;

	if (!json_object_is_type(values, json_type_object))
		return;
	if (!json_object_is_type(ifaces, json_type_array))
		return;

	if (strcmp(band, "all") != 0) {
		devicename = config_wireless_get_devicename(values, band);
		if (!devicename)
			return;
	}

	json_object_object_foreach(values, key, val) {

		type = json_get_string(val, ".type");
		if (!type || strcmp(type, "wifi-iface") != 0)
			continue;

		disabled = json_get_string(val, "disabled");
		if (disabled && strcmp(disabled, "1") == 0)
			continue;

		mode = json_get_string(val, "mode");
		if (mode && strcmp(mode, "ap") != 0)
			continue;

		net = json_get_string(val, "network");
		if (net && network && strcmp(net, network) != 0)
			continue;

		device = json_get_string(val, "device");
		if (!device)
			continue;
		if (devicename && strcmp(device, devicename) != 0)
			continue;

		band_string = config_wireless_get_band(values, device);
		if (!band_string)
			continue;
		band_option = json_object_new_string(band_string);
		if (!band_option)
			continue;
		json_object_object_add(val, "band", band_option);
		/* use json_object_get to increase the ref count of val */
		val = json_object_get(val);
		json_object_array_add(ifaces, val);
	}
}

/* get the first radio's band
 * should be used when only when radio count is 1
 */
static const char *config_get_single_band(struct json_object *values)
{
	int rv;
	const char *type;

	if (!json_object_is_type(values, json_type_object))
		goto fail;

	json_object_object_foreach(values, key, val) {
		type = json_get_string(val, ".type");
		if (!type || strcmp(type, "wifi-device") != 0)
			continue;

		return json_get_string(val, "band");
	}

fail:
	return NULL;
}

int config_ignore_option(const char *option)
{
	struct ignored_option *ign;

	ign = (struct ignored_option *)malloc(sizeof(*ign));
	if (!ign) {
		ERROR("malloc failed");
		return -1;
	}

	ign->option = strdup(option);
	if (!ign->option) {
		free(ign);
		ERROR("strdup failed");
		return -1;
	}

	list_add(&ign->list, &ignored_options);
	return 0;
}

void config_free_ignored_options(void)
{
	struct ignored_option *ign, *tmp;

	list_for_each_entry_safe(ign, tmp, &ignored_options, list) {
		list_del(&ign->list);
		free(ign->option);
		free(ign);
	}
}

/* return EQUAL if the options have the same type and value*/
static int config_options_equal(const void *obj1, const void *obj2)
{
	struct json_object *o1, *o2;
	enum json_type type1, type2;
	const char *str1, *str2;
	bool bool1, bool2;
	double d1, d2;
	int64_t int1, int2;
	int n1, n2, i, eq;

	o1 = (struct json_object *)obj1;
	o2 = (struct json_object *)obj2;
	/* is key right type? */
	type1 = json_object_get_type(o1);
	type2 = json_object_get_type(o2);
	if (type1 != type2)
		return type1 - type2;

	switch (type1) {
	case json_type_string:
		str1 = json_object_get_string(o1);
		str2 = json_object_get_string(o2);
		return strcmp(str1, str2);
	case json_type_boolean:
		bool1 = json_object_get_boolean(o1);
		bool2 = json_object_get_boolean(o2);
		return bool1 - bool2;
	case json_type_double:
		d1 = json_object_get_double(o1);
		d2 = json_object_get_double(o2);
		return d1 - d2;
	case json_type_int:
		int1 = json_object_get_int64(o1);
		int2 = json_object_get_int64(o2);
		return int1 - int2;
	case json_type_array:
		json_object_array_sort(o1, config_options_equal);
		json_object_array_sort(o2, config_options_equal);
		n1 = json_object_array_length(o1);
		n2 = json_object_array_length(o2);
		if (n1 != n2)
			return n1 - n2;
		for (i = 0; i < n1; i++) {
			eq = config_options_equal(
				json_object_array_get_idx(o1, i),
				json_object_array_get_idx(o2, i));
			if (eq != EQUAL)
				return eq;
		}
		return EQUAL;
	case json_type_null:
	case json_type_object:
	default:
		ERROR("ivalid type for o1 and o2: %s",
				json_type_to_name(type1));
	}
	return -1;
}

static int config_compare_bands(struct json_object *b1, struct json_object *b2)
{
	const char *band1, *band2;

	if (!json_object_is_type(b1, json_type_string)) {
		ERROR("band must be json_type_string");
		return !EQUAL;
	}
	if (!json_object_is_type(b2, json_type_string)) {
		ERROR("band must be json_type_string");
		return !EQUAL;
	}

	band1 = json_object_get_string(b1);
	band2 = json_object_get_string(b2);

	if (strlen(band1) == 0 || strlen(band2) == 0) {
		ERROR("band is empty: band1 = \"%s\" band2 = \"%s\"",
				band1, band2);
		return !EQUAL;
	}

	if (strcmp(band1, band2) == 0)
		return EQUAL;

	if (strcmp(band1, "all") == 0)
		return EQUAL;

	if (strcmp(band2, "all") == 0)
		return EQUAL;

	return !EQUAL;
}

static bool config_find_match(struct json_object *iface,
		struct json_object *list)
{
	bool section_match;
	int i, n;
	struct json_object *tmp, *value;
	json_bool rv;

	if (!json_object_is_type(iface, json_type_object)) {
		ERROR("iface: invalid type should be json_type_object");
		return false;
	}
	if (!json_object_is_type(list, json_type_array)) {
		ERROR("list: invalid type should be json_type_array");
		return false;
	}

	n = json_object_array_length(list);
	/* for every iface in list */
	for (i = 0; i < n; i++) {
		tmp = json_object_array_get_idx(list, i);
		if (!tmp) {
			ERROR("couldn't get iface nr %d from list", i);
			continue;
		}
		/* start by assuming that the sections match */
		section_match = true;
		/* for option in iface */
		json_object_object_foreach(iface, key, val) {
			if (is_ignored_option(key))
				continue;
			/* get the key from the iface from the list */
			rv = json_object_object_get_ex(tmp, key, &value);
			/* does key exist? */
			if (!rv || !value) {
				section_match = false;
				break;
			}

			/* special handle for band
			 * it can be "all", this should match both "a" and "b"
			 */
			if (strcmp(key, "band") == 0) {
				if (config_compare_bands(val, value) != EQUAL) {
					section_match = false;
					break;
				}
			} else if (config_options_equal(val, value) != EQUAL) {
				section_match = false;
				break;
			}
		}
		if (section_match)
			return true;
	}
	return false;
}

static enum config_compare_status config_compare_ifaces(
		struct json_object *remote_ifaces,
		struct json_object *local_ifaces)
{
	int n, i;
	struct json_object *iface;

	if (!json_object_is_type(remote_ifaces, json_type_array))
		return CONFIG_COMPARE_FAIL;
	if (!json_object_is_type(local_ifaces, json_type_array))
		return CONFIG_COMPARE_FAIL;

	/* for iface in remote find match in local or return */
	n = json_object_array_length(remote_ifaces);
	for (i = 0; i < n; i++) {
		iface = json_object_array_get_idx(remote_ifaces, i);
		if (!iface)
			continue;

		if (!config_find_match(iface, local_ifaces)) {
			DEBUG(LOG_NOTICE, "remote iface has no match:\n%s", json_object_to_json_string(iface));
			return CONFIG_COMPARE_OPTION_CHANGE;
		}
	}

	/* for iface in local find match in remote or return */
	n = json_object_array_length(local_ifaces);
	for (i = 0; i < n; i++) {
		iface = json_object_array_get_idx(local_ifaces, i);
		if (!iface)
			continue;

		if (!config_find_match(iface, remote_ifaces)) {
			DEBUG(LOG_NOTICE, "local iface has no match:\n%s", json_object_to_json_string(iface));
			return CONFIG_COMPARE_OPTION_CHANGE;
		}
	}
	return CONFIG_COMPARE_EQUAL;
}

static struct uci_section *get_wet_section(struct uci_context *ctx,
		struct uci_package *pkg)
{

	struct uci_element *e;
	struct uci_section *section;
	const char *mode;

	/* get the wet iface section */
	uci_foreach_element(&pkg->sections, e) {
		section = uci_to_section(e);
		if (strcmp(section->type, "wifi-iface") != 0)
			continue;

		mode = uci_lookup_option_string(ctx, section, "mode");
		if (mode && strcmp(mode, "wet") == 0)
			return section;
	}

	return NULL;
}

static bool option_is_equal(const char *s1, const char *s2)
{
	if (!s1 && !s2)
		return true;
	if (!s1 || !s2)
		return false;
	return strcmp(s1, s2) == 0;
}

static const char *get_band_by_device(struct uci_context *ctx,
		struct uci_package *pkg, const char *dev)
{
	struct uci_element *e;
	struct uci_section *s;

	uci_foreach_element(&pkg->sections, e) {
		s = uci_to_section(e);
		if (strcmp(s->type, "wifi-device") != 0)
			continue;
		if (strcmp(e->name, dev) == 0)
			return uci_lookup_option_string(ctx, s, "band");
	}
	return NULL;
}

static enum config_compare_status config_compare_wet(
		struct json_object *remote_ifaces, const char *remote_bssid)
{
	int n, i, rv;
	struct json_object *iface;
	struct uci_context *ctx;
	struct uci_package *package;
	struct uci_section *section;
	struct uci_element *se;
	enum config_compare_status return_status = CONFIG_COMPARE_FAIL;
	const char *wet_ssid, *wet_key, *wet_encryption, *wet_bssid, wet;
	const char *remote_ssid, *remote_key, *remote_encryption;
	const char *wetdev, *wetband, *remote_band;
	char *errstr;
	bool found;

	/* remote_ifaces MUST be of type array */
	if (!json_object_is_type(remote_ifaces, json_type_array))
		goto out;

	/* Preparation: load uci context/config */
	ctx = uci_alloc_context();
	if (!ctx)
		goto out;

	rv = uci_load(ctx, "wireless", &package);
	if (rv != UCI_OK) {
		uci_get_errorstr(ctx, &errstr, NULL);
		ERROR("%s", errstr);
		free(errstr);
		goto out_ctx;
	}

	section = get_wet_section(ctx, package);
	if (!section)
		goto out_pkg;

	wet_ssid = uci_lookup_option_string(ctx, section, "ssid");
	wet_key = uci_lookup_option_string(ctx, section, "key");
	wet_encryption = uci_lookup_option_string(ctx, section, "encryption");
	wet_bssid = uci_lookup_option_string(ctx, section, "connect_bssid");

	/* get the device of the wet iface */
	wetdev = uci_lookup_option_string(ctx, section, "device");
	if (!wetdev)
		goto out_pkg;

	/* get the band of the wet device */
	wetband = get_band_by_device(ctx, package, wetdev);
	if (!wetband)
		goto out_pkg;

	/* get the first remote iface on band wetband */
	n = json_object_array_length(remote_ifaces);
	found = false;
	for (i = 0; i < n; i++) {
		iface = json_object_array_get_idx(remote_ifaces, i);
		if (!iface)
			continue;
		remote_band = json_get_string(iface, "band");
		if (strcmp(remote_band, wetband) == 0) {
			found = true;
			break;
		}
	}

	if (!found)
		goto out_pkg;

	remote_ssid = json_get_string(iface, "ssid");
	remote_key = json_get_string(iface, "key");
	remote_encryption = json_get_string(iface, "encryption");

	if (!option_is_equal(wet_ssid, remote_ssid)) {
		DEBUG(LOG_NOTICE, "ssid changed: wet_ssid %s remote_ssid %s", wet_ssid, remote_ssid);
		return_status = CONFIG_COMPARE_WETIF_CHANGE;
		goto out_pkg;
	}
	if (!option_is_equal(wet_key, remote_key)) {
		DEBUG(LOG_NOTICE, "key changed: wet_key %s remote_key %s", wet_key, remote_key);
		return_status = CONFIG_COMPARE_WETIF_CHANGE;
		goto out_pkg;
	}
	if (!option_is_equal(wet_encryption, remote_encryption)) {
		DEBUG(LOG_NOTICE, "encryption changed: wet_encryption %s remote_encryption %s", wet_encryption, remote_encryption);
		return_status = CONFIG_COMPARE_WETIF_CHANGE;
		goto out_pkg;
	}
	if (!option_is_equal(wet_bssid, remote_bssid)) {
		DEBUG(LOG_NOTICE, "bssid changed: wet_bssid %s remote_bssid %s", wet_bssid, remote_bssid);
		return_status = CONFIG_COMPARE_WETIF_CHANGE;
		goto out_pkg;
	}

	return_status = CONFIG_COMPARE_EQUAL;

out_pkg:
	rv = uci_unload(ctx, package);
	if (rv != UCI_OK) {
		uci_get_errorstr(ctx, &errstr, NULL);
		ERROR("%s", errstr);
		free(errstr);
	}
out_ctx:
	uci_free_context(ctx);
out:
	return return_status;
}

static void config_filter_exceptions(struct json_object *ifaces)
{
	int i, n, j;
	struct json_object *iface;

	if (!json_object_is_type(ifaces, json_type_array))
		return;

	n = json_object_array_length(ifaces);
	for (i = 0; i < n; i++) {
		iface = json_object_array_get_idx(ifaces, i);
		if (!iface) {
			ERROR("couldn't find %d interface in array", i);
			continue;
		}
		json_object_object_foreach(iface, key, val) {
			for (j = 0; j < ARRAY_SIZE(filtered_options); j++) {
				if (strcmp(filtered_options[j], key) == 0) {
					json_object_object_del(iface, key);
				}
			}

		}
	}
}

static void ifaces_on_all_bands(struct json_object *ifaces)
{
	int i, n;
	struct json_object *iface, *band;

	if (!json_object_is_type(ifaces, json_type_array))
		return;

	n = json_object_array_length(ifaces);
	for (i = 0; i < n; i++) {
		iface = json_object_array_get_idx(ifaces, i);
		if (!iface)
			continue;
		band = json_object_new_string("all");
		if (!band)
			continue;
		json_object_object_del(iface, "band");
		json_object_object_add(iface, "band", band);
	}
}

/* local* is the config on the current router
 * remote* is the received config
 */
enum config_compare_status config_compare_wireless(
	struct json_object *local_config, struct json_object *remote_config,
	struct json_object *remote_ifaces, const char *network,
	const char *remote_bssid)
{
	int local_nradio, remote_nradio;
	int local_niface, remote_niface;
	int remote_multiplier = 1;
	struct json_object *local_values, *remote_values, *local_ifaces = NULL;
	const char *band;
	json_bool jrv;
	enum config_compare_status rv = CONFIG_COMPARE_FAIL;

	local_ifaces = json_object_new_array();
	if (!local_ifaces) {
		goto out;
	}
	if (!json_object_is_type(remote_ifaces, json_type_array)) {
		ERROR("remote_ifaces is not json_type_array");
		goto out;
	}

	/* values */
	jrv = json_object_object_get_ex(local_config, "values", &local_values);
	if (!jrv) {
		ERROR("no values in local_config");
		goto out;
	}
	jrv = json_object_object_get_ex(remote_config, "values",
			&remote_values);
	if (!jrv) {
		ERROR("no values in remote_config");
		goto out;
	}

	local_nradio = config_count_radio(local_values);
	remote_nradio = config_count_radio(remote_values);
	if (local_nradio < 1 || remote_nradio < 1 ||
	    local_nradio > 2 || remote_nradio > 2) {
		ERROR("invalid number of radios (1-2): %d %d",
				local_nradio, remote_nradio);
		goto out;
	}

	/* if local has less radios than remote:
	 * count only the single band ifaces
	 */
	if (local_nradio < remote_nradio) {
		band = config_get_single_band(local_values);
		if (!band) {
			ERROR("could not find single band in local_values");
			goto out;
		}
		config_count_wifi_ifaces(local_values, band, local_ifaces, NULL);
		config_count_wifi_ifaces(remote_values, band, remote_ifaces, network);
	}

	/* if remote has less radios than local:
	 * double the remote ifaces
	 */
	if (local_nradio > remote_nradio) {
		config_count_wifi_ifaces(local_values, "all", local_ifaces, NULL);
		config_count_wifi_ifaces(remote_values, "all", remote_ifaces, network);
		ifaces_on_all_bands(remote_ifaces);
		remote_multiplier = 2;
	}

	/* if remote the same number of radios as local */
	if (local_nradio == remote_nradio) {
		config_count_wifi_ifaces(local_values, "all", local_ifaces, NULL);
		config_count_wifi_ifaces(remote_values, "all", remote_ifaces, network);
	}

	local_niface = json_object_array_length(local_ifaces);
	remote_niface = json_object_array_length(remote_ifaces);
	remote_niface *= remote_multiplier;
	DEBUG(LOG_DEBUG, "local_niface = %d", local_niface);
	DEBUG(LOG_DEBUG, "remote_niface = %d", remote_niface);

	/* compare the (adjusted) iface amounts (local and remote)
	 * if they are different, return SECTION_CHANGE
	 * (then proceed with APPLYing the new config)
	 * else decide based on the iface content (look deeper)
	 */
	config_filter_exceptions(remote_ifaces);
	config_filter_exceptions(local_ifaces);

	DEBUG(LOG_DEBUG, "remote_ifaces:\n%s\n", json_object_to_json_string_ext(remote_ifaces, JSON_C_TO_STRING_PRETTY));
	DEBUG(LOG_DEBUG, "local_ifaces:\n%s\n", json_object_to_json_string_ext(local_ifaces, JSON_C_TO_STRING_PRETTY));

	if (local_niface != remote_niface) {
		rv = CONFIG_COMPARE_SECTION_CHANGE;
		goto out;
	}

	rv = config_compare_ifaces(remote_ifaces, local_ifaces);

	if (rv != CONFIG_COMPARE_EQUAL)
		goto out;

	rv = config_compare_wet(remote_ifaces, remote_bssid);

out:
	DEBUG(LOG_NOTICE, "compare returned with result %d", rv);
	if (local_ifaces)
		json_object_put(local_ifaces);
	return rv;
}

static int set_value(struct uci_context *ctx, struct uci_package *pkg,
		struct uci_section *section, const char *key,
		const char *value, enum uci_option_type type)
{
	struct uci_ptr ptr = {0};

	ptr.p = pkg;
	ptr.s = section;
	ptr.option = key;
	ptr.value = value;
	if (type == UCI_TYPE_STRING)
		return uci_set(ctx, &ptr);
	if (type == UCI_TYPE_LIST)
		return uci_add_list(ctx, &ptr);
	return -1;
}

/* ptr must have the correct section set in ptr->s */
static int set_network(struct uci_context *ctx, struct uci_package *pkg,
		struct uci_section *section)
{
	const char *network;

	// TODO: find out network!!!
	network = "wan";
	return set_value(ctx, pkg, section, "network", network,
			UCI_TYPE_STRING);
}

static const char *get_device_by_band(struct uci_context *ctx,
		struct uci_package *pkg, const char *b)
{
	struct uci_element *e;
	struct uci_section *s;
	const char *band;

	uci_foreach_element(&pkg->sections, e) {
		s = uci_to_section(e);
		if (strcmp(s->type, "wifi-device") != 0)
			continue;
		band = uci_lookup_option_string(ctx, s, "band");
		if (strcmp(band, b) == 0)
			return e->name;
	}
	return NULL;
}

static bool create_new_wifi_iface(struct uci_context *ctx,
		struct uci_package *pkg, const char *b)
{
	struct uci_section *s = NULL;
	struct uci_ptr ptr;
	const char *device, *network;
	int ret, rv;

	device = get_device_by_band(ctx, pkg, b);
	/* if there is no device with band b dont create the interface */
	if (!device)
		return true;

	rv = uci_add_section(ctx, pkg, "wifi-iface", &s);
	if (rv != 0)
		return false;

	memset(&ptr, 0, sizeof(ptr));
	ptr.p = pkg;
	ptr.s = s;
	ret = set_network(ctx, ptr.p, ptr.s);

	return ret == 0;
}

static void apply_iface_config(struct uci_context *ctx, struct uci_package *pkg,
		struct json_object *iface, const char *band, int *iface_num,
		bool ignore_exceptions)
{
	const char *device, *mode;
	struct uci_element *e;
	struct uci_section *section = NULL, *tmp;
	struct json_object *obj;
	int i = *iface_num, k, len;

	device = get_device_by_band(ctx, pkg, band);
	if (!device)
		return;

	uci_foreach_element(&pkg->sections, e) {
		tmp = uci_to_section(e);
		if (strcmp(tmp->type, "wifi-iface") != 0)
			continue;

		mode = uci_lookup_option_string(ctx, tmp, "mode");
		if (mode && strcmp(mode, "ap") != 0)
			continue;
		if (i == 0) {
			section = tmp;
			break;
		}
		i--;
	}
	if (!section)
		return;

	set_value(ctx, pkg, section, "device", device, UCI_TYPE_STRING);
	json_object_object_foreach(iface, key, val) {
		if (ignore_exceptions && is_ignored_option(key))
			continue;
		switch (json_object_get_type(val)) {
		case json_type_string:
			if (strcmp(key, "band") == 0)
				break;
			set_value(ctx, pkg, section, key,
					json_object_get_string(val),
					UCI_TYPE_STRING);
			break;
		case json_type_array:
			len = json_object_array_length(val);
			for (k = 0; k < len; k++) {
				obj = json_object_array_get_idx(val, k);
				if (!json_object_is_type(obj, json_type_string))
					continue;
				set_value(ctx, pkg, section, key,
						json_object_get_string(obj),
						UCI_TYPE_LIST);
			}
			break;
		default:
			break;
		}
	}

	(*iface_num)++;
}

static void apply_wet_iface(struct uci_context *ctx, struct uci_package *pkg,
		struct json_object *remote_ifaces, const char *remote_bssid)
{
	const char *mode, *wetdev, *wetband, *remote_band;
	const char *ssid, *key, *encryption, *bssid;
	struct uci_section *section;
	struct json_object *iface;
	int i, num_ifaces;
	struct uci_ptr ptr = {0};

	if (!json_object_is_type(remote_ifaces, json_type_array)) {
		ERROR("remote_ifaces is not json_type_array");
		return;
	}

	section = get_wet_section(ctx, pkg);
	if (!section)
		return;

	/* get the device of the wet iface */
	wetdev = uci_lookup_option_string(ctx, section, "device");
	if (!wetdev)
		return;

	/* get the band of the wet device */
	wetband = get_band_by_device(ctx, pkg, wetdev);
	if (!wetband)
		return;

	/* get the first remote iface on band wetband */
	num_ifaces = json_object_array_length(remote_ifaces);
	ssid = NULL;
	for (i = 0; i < num_ifaces; i++) {
		iface = json_object_array_get_idx(remote_ifaces, i);
		remote_band = json_get_string(iface, "band");
		if (!remote_band)
			continue;

		if (strcmp(wetband, remote_band) != 0)
			continue;

		ssid = json_get_string(iface, "ssid");
		if (!ssid)
			continue;
		/* found it */
		break;
	}
	if (!ssid)
		return;

	key = json_get_string(iface, "key");
	encryption = json_get_string(iface, "encryption");

	/* apply the remote iface on local wet iface */
	set_value(ctx, pkg, section, "ssid", ssid, UCI_TYPE_STRING);
	if (key)
		set_value(ctx, pkg, section, "key", key, UCI_TYPE_STRING);
	if (encryption)
		set_value(ctx, pkg, section,
			"encryption", encryption, UCI_TYPE_STRING);
	if (remote_bssid)
		set_value(ctx, pkg, section,
			"connect_bssid", remote_bssid, UCI_TYPE_STRING);
	else {
		/* if connect connect_bssid is not present then we want to remove it
		to enable daisy chaining */
		ptr.p = pkg;
		ptr.s = section;
		ptr.option = "connect_bssid";
		uci_delete(ctx, &ptr);
	}
}

struct json_object *config_get_section(struct uci_context *ctx, struct uci_section *section, struct json_object *cfg)
{
	const char *type, *name, *cfg_name;

	json_object_object_foreach(cfg, key, val) {
		type = json_get_string(val, ".type");
		if (!type)
			continue;

		if (strncmp(type, section->type, 128))
			continue;

		if (!strncmp(type, "steer-param", strlen("steer-param"))) {
			name = json_get_string(val, "type");
			if (!name)
				continue;

			cfg_name = uci_lookup_option_string(ctx, section, "type");
			if (!cfg_name)
				continue;

			if (strncmp(name, cfg_name, 128))
				continue;
		}

		return val;
	}

	return NULL;
}

enum config_apply_status config_apply_wifilife(struct json_object *cfg)
{
	struct uci_context *ctx;
	struct uci_package *package;
	struct uci_section *section;
	struct uci_option *option;
	struct uci_ptr ptr;
	struct uci_element *se, *tmp;
	struct json_object *elem, *json_section;
	enum config_apply_status return_status = CONFIG_APPLY_FAIL;
	char *errstr;
	int rv, i;

	ctx = uci_alloc_context();
	if (!ctx)
		goto out;

	rv = uci_load(ctx, "wifilife", &package);
	if (rv != UCI_OK) {
		uci_get_errorstr(ctx, &errstr, NULL);
		ERROR("%s", errstr);
		free(errstr);
		goto out_ctx;
	}

	uci_foreach_element_safe(&package->sections, tmp, se) {
		section = uci_to_section(se);
		if (!section)
			continue;

		json_section = config_get_section(ctx, section, cfg);
		if (!json_section)
			continue;

		json_object_object_foreach(json_section, key, val) {
			if (key[0] == '.')
				continue;

			if (!strncmp(key, "ifname", strlen("ifname")))
				continue;

			if (!strncmp(key, "exclude", strlen("exclude")))
				continue;

			if (json_object_is_type(val, json_type_array)) {
				option = uci_lookup_option(ctx, section, key);
				if (option) {
					memset(&ptr, 0, sizeof(ptr));
					ptr.p = package;
					ptr.s = section;
					ptr.o = option;
					uci_delete(ctx, &ptr);
				}

				for (i = 0; i < json_object_array_length(val); i++) {
					elem = json_object_array_get_idx(val, i);
					rv = set_value(ctx, package, section, key, json_object_get_string(elem), UCI_TYPE_LIST);
					if (rv != UCI_OK)
						break;
				}
			} else if (json_object_is_type(val, json_type_string))
				rv = set_value(ctx, package, section, key, json_object_get_string(val), UCI_TYPE_STRING);
			else
				continue;

			if (rv != UCI_OK) {
				uci_get_errorstr(ctx, &errstr, NULL);
				free(errstr);
				goto out_pkg;
			}
		}
	}

	return_status = CONFIG_APPLY_SUCCESS;
	uci_commit(ctx, &package, false);
out_pkg:
	uci_unload(ctx, package);
out_ctx:
	uci_free_context(ctx);
out:
	return return_status;
}


enum config_apply_status config_apply_wireless(
	struct json_object *remote_ifaces, const char *remote_bssid,
	enum config_compare_status status)
{
	struct uci_context *ctx;
	struct uci_package *package;
	struct uci_section *section;
	struct uci_option *option;
	struct uci_ptr ptr;
	struct uci_element *se, *oe, *tmp;
	struct json_object *iface;
	enum config_apply_status return_status = CONFIG_APPLY_FAIL;
	const char *mode, *b, network;
	char *errstr;
	int rv, i, num_ifaces, iface_num;
	bool ignore_exceptions;

	/* remote_ifaces MUST be of type array */
	if (!json_object_is_type(remote_ifaces, json_type_array))
		goto out;

	if (status != CONFIG_COMPARE_SECTION_CHANGE &&
	    status != CONFIG_COMPARE_OPTION_CHANGE &&
	    status != CONFIG_COMPARE_WETIF_CHANGE)
		goto out;

	/* Preparation: load uci context/config */
	ctx = uci_alloc_context();
	if (!ctx)
		goto out;

	rv = uci_load(ctx, "wireless", &package);
	if (rv != UCI_OK) {
		uci_get_errorstr(ctx, &errstr, NULL);
		ERROR("%s", errstr);
		free(errstr);
		goto out_ctx;
	}
	/* step 1: apply the wet iface */
	apply_wet_iface(ctx, package, remote_ifaces, remote_bssid);

	if (status == CONFIG_COMPARE_WETIF_CHANGE) {
		return_status = CONFIG_APPLY_SUCCESS;
		goto out_commit;
	}

	/* step 2: clean the current config */

	if (status == CONFIG_COMPARE_SECTION_CHANGE) {
		/* 1. delete all local ifaces from uci config */
		uci_foreach_element_safe(&package->sections, tmp, se) {
			section = uci_to_section(se);
			/* only delete wifi-iface sections */
			if (strcmp(section->type, "wifi-iface") != 0)
				continue;
			/* delete only AP interfaces (mode == ap or not set) */
			mode = uci_lookup_option_string(ctx, section, "mode");
			if (mode && strcmp(mode, "ap") != 0)
				continue;
			memset(&ptr, 0, sizeof(ptr));
			ptr.p = package;
			ptr.s = section;
			uci_delete(ctx, &ptr);
		}
		/** 2. create new interfaces according to the remote config
		 * (for ifaces with band == "all", create them on each radio)
		 * and generate network and device
		 **/
		num_ifaces = json_object_array_length(remote_ifaces);
		for (i = 0; i < num_ifaces; i++) {
			iface = json_object_array_get_idx(remote_ifaces, i);

			b = json_get_string(iface, "band");
			if (!b)
				goto out_pkg;
			if (strcmp(b, "a") == 0 || strcmp(b, "all") == 0) {
				if (!create_new_wifi_iface(ctx, package, "a")) {
					ERROR("unable to create uci section");
					goto out_pkg;
				}
			}
			if (strcmp(b, "b") == 0 || strcmp(b, "all") == 0) {
				if (!create_new_wifi_iface(ctx, package, "b")) {
					ERROR("unable to create uci section");
					goto out_pkg;
				}
			}
		}
	}

	if (status == CONFIG_COMPARE_OPTION_CHANGE) {
		/** 1. keep only device name
		 * and values added from cmd line with --ignore|-i
		 * 2. set network on all sections
		 **/
		uci_foreach_element(&package->sections, se) {
			section = uci_to_section(se);
			if (strcmp(section->type, "wifi-iface") != 0)
				continue;
			/* clean only AP interfaces (mode == ap or not set) */
			mode = uci_lookup_option_string(ctx, section, "mode");
			if (mode && strcmp(mode, "ap") != 0)
				continue;
			uci_foreach_element_safe(&section->options, tmp, oe) {
				if (strcmp(oe->name, "network") == 0) {
					set_network(ctx, package, section);
					continue;
				}

				if (is_ignored_option(oe->name))
					continue;
				/* delete value */
				memset(&ptr, 0, sizeof(ptr));
				ptr.p = package;
				ptr.o = uci_to_option(oe);
				ptr.s = section;
				uci_delete(ctx, &ptr);
			}
		}
	}
	/** step 3: apply the new config
	 * foreach remote iface
	 *	if (remote has band a)
	 *		apply iface on local band a
	 *	if (remote has band b)
	 *		apply iface on local band b
	 **/
	num_ifaces = json_object_array_length(remote_ifaces);
	iface_num = 0;
	ignore_exceptions = (status == CONFIG_COMPARE_OPTION_CHANGE);
	for (i = 0; i < num_ifaces; i++) {
		iface = json_object_array_get_idx(remote_ifaces, i);
		b = json_get_string(iface, "band");
		if (!b)
			goto out_pkg;
		// apply for band a
		if (strcmp(b, "a") == 0 || strcmp(b, "all") == 0)
			apply_iface_config(ctx, package, iface, "a", &iface_num,
					ignore_exceptions);
		// apply for band b
		if (strcmp(b, "b") == 0 || strcmp(b, "all") == 0)
			apply_iface_config(ctx, package, iface, "b", &iface_num,
					ignore_exceptions);
	}

out_commit:
	/* save config */
	uci_commit(ctx, &package, false);
out_pkg:
	uci_unload(ctx, package);
	return_status = CONFIG_APPLY_SUCCESS;
out_ctx:
	uci_free_context(ctx);
out:
	return return_status;
}

static bool same_network(const char *ip1_str, const char *ip2_str,
		const char *mask_str)
{
	struct in_addr ip1, ip2, mask;
	int rv;
	//maybe TODO: IPv6?

	rv  = inet_pton(AF_INET, ip1_str, &ip1);
	rv += inet_pton(AF_INET, ip2_str, &ip2);
	rv += inet_pton(AF_INET, mask_str, &mask);
	/* inet_pton returns 1 on success so all success means 3 * 1 = 3 */
	if (rv != 3)
		return false;
	return (ip1.s_addr & mask.s_addr) == (ip2.s_addr & mask.s_addr);
}

char *config_get_network_from_ip(const char *ip)
{
	char *network = NULL;
	struct uci_context *ctx;
	struct uci_element *se;
	struct uci_package *package;
	struct uci_section *section;
	const char *is_lan, *ip_addr, *mask;
	char *errstr;
	int rv;

	ctx = uci_alloc_context();
	if (!ctx)
		goto out;

	rv = uci_load(ctx, "network", &package);
	if (rv != UCI_OK) {
		uci_get_errorstr(ctx, &errstr, NULL);
		ERROR("%s", errstr);
		free(errstr);
		goto out_ctx;
	}

	/* for each section */
	uci_foreach_element(&package->sections, se) {
		section = uci_to_section(se);
		if (strcmp(section->type, "interface") != 0)
			continue;

		ip_addr = uci_lookup_option_string(ctx, section, "ipaddr");
		if (!ip_addr)
			continue;
		mask = uci_lookup_option_string(ctx, section, "netmask");
		if (!ip_addr)
			continue;

		if (!same_network(ip, ip_addr, mask))
			continue;
		network = strdup(se->name);
		break;
	}
	rv = uci_unload(ctx, package);
	if (rv != UCI_OK) {
		uci_get_errorstr(ctx, &errstr, NULL);
		ERROR("%s", errstr);
		free(errstr);
	}
out_ctx:
	uci_free_context(ctx);
out:
	return network;
}

char *config_get_ifname_by_band(char *ifname, char *band)
{
	struct uci_context *ctx;
	struct uci_element *se;
	struct uci_package *package;
	struct uci_section *section;
	const char *is_lan, *ip_addr, *mask;
	char *errstr;
	const char *device = NULL;
	int rv;

	ctx = uci_alloc_context();
	if (!ctx)
		goto out;

	rv = uci_load(ctx, "wireless", &package);
	if (rv != UCI_OK) {
		uci_get_errorstr(ctx, &errstr, NULL);
		ERROR("%s", errstr);
		free(errstr);
		goto out_ctx;
	}

	device = get_device_by_band(ctx, package, band);
	if (!device)
		goto out_unload;

	strncpy(ifname, device, IFNAMSIZ);

out_unload:
	rv = uci_unload(ctx, package);
	if (rv != UCI_OK) {
		uci_get_errorstr(ctx, &errstr, NULL);
		ERROR("%s", errstr);
		free(errstr);
	}
out_ctx:
	uci_free_context(ctx);
out:
	return ifname;
}

int config_get_radios(struct list_head *radios)
{
	int rv = -1, section_index = 0;
	char *errstr;
	struct uci_context *ctx;
	struct uci_element *se, *oe, *le;
	struct uci_package *package;
	struct uci_section *section;
	struct uci_option *option;
	void *vt, *ot, *la;
	struct radio *r, *tmp;

	/* Make sure list is empty */
	list_for_each_entry_safe(r, tmp, radios, list) {
		DEBUG(LOG_DEBUG, "New credential exchange initiated, cleaning old \
				entry %s from the radio list", r->ifname);
		list_del(&r->list);
		free(r);
	}

	ctx = uci_alloc_context();
	if (!ctx)
		goto out;

	rv = uci_load(ctx, "wireless", &package);
	if (rv != UCI_OK) {
		uci_get_errorstr(ctx, &errstr, NULL);
		ERROR("%s", errstr);
		free(errstr);
		goto out_ctx;
	}

	uci_foreach_element(&package->sections, se) {
		section = uci_to_section(se);
		if (!section)
			continue;

		if (strncmp(section->type, "wifi-device", strlen("wifi-device")) != 0)
			continue;

		r = calloc(1, sizeof(*r));
		if (!r)
			break;

		DEBUG(LOG_DEBUG, "Added radio %s to the radio list", section->e.name);

		strncpy(r->ifname, section->e.name, IFNAMSIZ - 1);
		list_add(&r->list, radios);
	}

	rv = uci_unload(ctx, package);
	if (rv != UCI_OK) {
		uci_get_errorstr(ctx, &errstr, NULL);
		ERROR("%s", errstr);
		free(errstr);
	}

out_ctx:
	uci_free_context(ctx);
out:
	return rv;
}
