/**
 * netmoded -- netmode daemon for iopsysWrt powered CPEs
 *
 * Copyright (C) 2018-2019 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: Reidar Cederqvist <reidar.cederqvist@gmail.com>
 * Author: Ionut-Alex Oprea <ionutalexoprea@gmail.com>
 * Author: Jakob Olsson <jakob.olsson@iopsys.eu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 **/

#include <arpa/inet.h>
#include <libubox/uloop.h>
#include <libubox/blobmsg_json.h>
#include <libubus.h>
#include <json-c/json.h>
#include <getopt.h>
#include <pthread.h>

#include <sys/socket.h>
#include <sys/ioctl.h>
#include <linux/if.h>
#include <netdb.h>

#include "netmoded.h"
#include "config.h"

#define NETMODE_UBUS_NAME "netmode"
#define MAX_IFACES 8

struct ubus_context *ctx;
struct ubus_event_handler ev;
struct uloop_timeout flash_timeout;
struct list_head radios;
int LOG_LEVEL = LOG_ERROR;

static int netmode_status(struct ubus_context *ctx, struct ubus_object *obj,
		struct ubus_request_data *req, const char *method,
		struct blob_attr *msg);
static int netmode_repeat(struct ubus_context *ctx, struct ubus_object *obj,
		struct ubus_request_data *req, const char *method,
		struct blob_attr *msg);
static int netmode_sync(struct ubus_context *ctx, struct ubus_object *obj,
		struct ubus_request_data *req, const char *method,
		struct blob_attr *msg);
static int netmode_repeat_wifilife(struct ubus_context *ctx, struct ubus_object *obj,
		struct ubus_request_data *req, const char *method,
		struct blob_attr *msg);
static int netmode_sync_wifilife(struct ubus_context *ctx, struct ubus_object *obj,
		struct ubus_request_data *req, const char *method,
		struct blob_attr *msg);

/* ubus object */
enum {
	P_NETWORK,
	P_WIRELESS,
	P_CONNECT_BSSID
};
struct blobmsg_policy netmode_repeat_policy[] = {
	[P_NETWORK] = {.name = "network", .type = BLOBMSG_TYPE_STRING},
	[P_WIRELESS] = {.name = "wireless", .type = BLOBMSG_TYPE_TABLE},
	[P_CONNECT_BSSID] = {.name = "connect_bssid", .type = BLOBMSG_TYPE_STRING},
};

enum {
	P_WIFILIFE
};

struct blobmsg_policy netmode_repeat_wifilife_policy[] = {
	[P_WIFILIFE] = {.name = "wifilife", .type = BLOBMSG_TYPE_TABLE},
};

struct ubus_method netmode_methods[] = {
//	UBUS_METHOD_NOARG("status", netmode_status),
	UBUS_METHOD("repeat", netmode_repeat, netmode_repeat_policy),
	UBUS_METHOD("repeat_wifilife", netmode_repeat_wifilife,
				netmode_repeat_wifilife_policy),
	UBUS_METHOD_NOARG("sync", netmode_sync),
	UBUS_METHOD_NOARG("sync_wifilife", netmode_sync_wifilife)};

struct ubus_object netmode_object =
		NETMODE_UBUS_OBJECT(NETMODE_UBUS_NAME, netmode_methods);

struct uloop_timeout sync_timer = {};
pthread_mutex_t sync_mutex = PTHREAD_MUTEX_INITIALIZER;

static int ubus_init(void)
{
	uloop_init();
	ctx = ubus_connect(NULL);
	if (!ctx)
		return -1;
	ubus_add_uloop(ctx);
	return 0;
}

static char *ifname_to_hwaddr(char *key, char *mac)
{
	int fd, i, rv;
	struct ifreq ifr;
	unsigned char *buffer;

	fd = socket(AF_INET, SOCK_DGRAM, 0);
	ifr.ifr_addr.sa_family = AF_INET;
	strncpy(ifr.ifr_name, key, IFNAMSIZ);

	rv = ioctl(fd, SIOCGIFHWADDR, &ifr);
	if (rv < 0)
		goto fail_ioctl;

	buffer = (unsigned char *)ifr.ifr_hwaddr.sa_data;

	snprintf(mac, 18, "%.2x:%.2x:%.2x:%.2x:%.2x:%.2x", buffer[0], buffer[1], buffer[2], buffer[3], buffer[4], buffer[5]);

	close(fd);
	return mac;
fail_ioctl:
	close(fd);
	return NULL;
}

static bool is_real_ip(const char *ip_str)
{
	struct in6_addr ip6;
	struct in_addr ip;
	int rv;

	rv = inet_pton(AF_INET, ip_str, &ip);
	if (rv)
		return true;

	rv = inet_pton(AF_INET6, ip_str, &ip6);
	return !!rv;
}

static bool is_remote_netmode_object(const char *object)
{
	char *slash, *ip_str;
	int ip_len;
	bool rv = false;

	slash = strchr(object, '/');
	if (!slash)
		goto out;

	if (strncmp(slash + 1, NETMODE_UBUS_NAME, strlen(NETMODE_UBUS_NAME))
			!= 0)
		goto out;

	ip_len = slash - object;
	ip_str = (char *)malloc((ip_len + 1) * sizeof(char));
	if (!ip_str) {
		ERROR("malloc failed");
		goto out;
	}

	memcpy(ip_str, object, ip_len);
	ip_str[ip_len] = '\0';

	rv = is_real_ip(ip_str);
	free(ip_str);
out:
	return rv;
}

/* TODO: Extract some portions into separate functions for re-usability */
static void send_wifilife_credentials(int remote_id, const char *path)
{
	int rv;
	struct blob_buf buf;
	void *wifilife_t;
	const char *slash;
	size_t size;
	char *ip, *network;

	memset(&buf, 0, sizeof(buf));
	blob_buf_init(&buf, 0);

	slash = strchr(path, '/');

	if (!slash)
		goto out;

	size = slash - path;
	ip = (char *)malloc(size + 1);
	if (!ip) {
		ERROR("couldn't malloc ip");
		goto out;
	}

	memcpy(ip, path, size);
	ip[size] = '\0';
	if (!is_real_ip(ip))
		goto out_ip;

	wifilife_t = blobmsg_open_table(&buf, "wifilife");
	rv = config_get_wifilife(&buf);
	if (rv)
		goto out_ip;

	blobmsg_close_table(&buf, wifilife_t);

	DEBUG(LOG_DEBUG, "attempting to send wifilife credentials");

	rv = ubus_invoke(ctx, remote_id, "repeat_wifilife", buf.head, NULL, NULL, 5000);
	if (rv != UBUS_STATUS_OK)
		ERROR("ubus_invoke <ip>/netmode repeat failed");

out_ip:
	free(ip);
out:
	blob_buf_free(&buf);
}

static void add_connect_bssid(struct blob_buf *bb)
{
	char bssid[18], ifname[16];

	/* band 'a' is 5GHz */
	config_get_ifname_by_band(ifname, "a");
	if (!ifname)
		return;

	if (!ifname_to_hwaddr(ifname, bssid))
		return;

	blobmsg_add_string(bb, "connect_bssid", bssid);
}

static void	send_credentials(int remote_id, const char *path)
{
	int rv;
	struct blob_buf buf;
	void *wireless_t;
	const char *slash;
	size_t size;
	char *ip, *network;

	memset(&buf, 0, sizeof(buf));
	blob_buf_init(&buf, 0);

	slash = strchr(path, '/');

	if (!slash)
		goto out;

	size = slash - path;
	ip = (char *)malloc(size + 1);
	if (!ip) {
		ERROR("couldn't malloc ip");
		goto out;
	}

	memcpy(ip, path, size);
	ip[size] = '\0';
	if (!is_real_ip(ip))
		goto out_ip;

	/* TODO: fix this: it only works if the master is the network gateway */
	network = config_get_network_from_ip(ip);
	DEBUG(LOG_DEBUG, "network %s", network);
	if (!network)
		goto out_ip;

	blobmsg_add_string(&buf, "network", network);
	wireless_t = blobmsg_open_table(&buf, "wireless");
	rv = config_get_wireless(&buf);
	blobmsg_close_table(&buf, wireless_t);

	if (!config_get_daisy_chain_enabled())
		add_connect_bssid(&buf);

	if (rv)
		goto out_network;

	rv = ubus_invoke(ctx, remote_id, "repeat", buf.head, NULL, NULL, 5000);
	if (rv != UBUS_STATUS_OK)
		ERROR("ubus_invoke <ip>/netmode repeat failed");

out_network:
	free(network);
out_ip:
	free(ip);
out:
	blob_buf_free(&buf);
}

void initial_send_wifilife_credentials_cb(struct ubus_context *ctx,
		struct ubus_object_data *obj, void *priv)
{
	if (is_remote_netmode_object(obj->path))
		send_wifilife_credentials(obj->id, obj->path);
}

void initial_send_credentials_cb(struct ubus_context *ctx,
		struct ubus_object_data *obj, void *priv)
{
	if (is_remote_netmode_object(obj->path))
		send_credentials(obj->id, obj->path);
}

static void initial_send_credentials_wifilife(struct uloop_timeout *timer)
{

	ubus_lookup(ctx, "*", initial_send_wifilife_credentials_cb, NULL);

	if (timer == &sync_timer)
		pthread_mutex_unlock(&sync_mutex);
}

static void initial_send_credentials(struct uloop_timeout *timer)
{

	ubus_lookup(ctx, "*", initial_send_credentials_cb, NULL);

	if (timer == &sync_timer)
		pthread_mutex_unlock(&sync_mutex);
}

static void handle_ubus_object_add_event(struct ubus_context *ctx,
		struct ubus_event_handler *ev, const char *type,
		struct blob_attr *msg)
{
	const char *msg_str, *path_str;
	json_object *msg_obj = NULL, *path_obj, *id_obj;
	json_bool rv;
	int id;

	msg_str = blobmsg_format_json(msg, true);
	if (!msg_str)
		goto out;

	msg_obj = json_tokener_parse(msg_str);
	if (!msg_obj)
		goto out;

	rv = json_object_object_get_ex(msg_obj, "path", &path_obj);
	if (!rv || !path_obj ||
			!json_object_is_type(path_obj, json_type_string))
		goto clean_msg_obj;

	path_str = json_object_get_string(path_obj);
	if (!path_str)
		goto clean_msg_obj;

	if (!is_remote_netmode_object(path_str))
		goto clean_msg_obj;

	DEBUG(LOG_NOTICE, "got event\n\"%s\"", blobmsg_format_json(msg, true));

	rv = json_object_object_get_ex(msg_obj, "id", &id_obj);
	if (!rv || !id_obj) {
		ERROR("json_object_object_get_ex %d %d", !rv, !id_obj);
		goto clean_msg_obj;
	}

	if (!id_obj || !json_object_is_type(id_obj, json_type_int)) {
		ERROR("json_object_is_type failed");
		goto clean_msg_obj;
	}

	errno = 0;
	id = json_object_get_int(id_obj);
	if (!id && errno == EINVAL) {
		ERROR("json_object_get_int id failed");
		goto clean_msg_obj;
	}

	send_credentials(id, path_str);
	send_wifilife_credentials(id, path_str);
clean_msg_obj:
	json_object_put(msg_obj);
out:
	return;
}

static int create_listener(struct ubus_event_handler *handler)
{
	handler->cb = handle_ubus_object_add_event;
	return ubus_register_event_handler(ctx, handler, "ubus.object.add");
}

void start_initial_timer_wifilife(struct uloop_timeout *timer)
{
	timer->cb = initial_send_credentials_wifilife;
	uloop_timeout_set(timer, 0);
	uloop_timeout_add(timer);
}

void start_initial_timer(struct uloop_timeout *timer)
{
	timer->cb = initial_send_credentials;
	uloop_timeout_set(timer, 0);
	uloop_timeout_add(timer);
}

int set_led_state(char *state)
{
	int obj, rv = -1;
	struct blob_buf buf;

	if (!ctx) {
		ERROR("No ubus context allocated");
		goto out;
	}

	ubus_lookup_id(ctx, "leds", &obj);
	if (!obj) {
		ERROR("Couldn't lookup the id of ubus leds object");
		goto out;
	}

	memset(&buf, 0, sizeof(buf));
	rv = blob_buf_init(&buf, 0);
	if (rv) {
		ERROR("Couldn't initialize blob_buf");
		goto out;
	}

	rv = blobmsg_add_string(&buf, "state", state);
	if (rv) {
		ERROR("Couldn't add string to blob_buf");
		goto out_buf;
	}

	rv = ubus_invoke(ctx, obj, "set", buf.head, NULL, NULL, 5000);
	if (rv)
		ERROR("Couldn't run ubus call leds set %s",
				blobmsg_format_json(buf.head, true));

out_buf:
	blob_buf_free(&buf);
out:
	return rv;
}

static void stop_flash_cb(struct uloop_timeout *t)
{
	struct radio *r, *tmp;

	ubus_unregister_event_handler(ctx, &ev);
	set_led_state("normal");

	/* if we reach time limit, free any remaining entries */
	list_for_each_entry_safe(r, tmp, &radios, list) {
		DEBUG(LOG_DEBUG, "Time limit reached, remove leftover entry %s \
				from radio list", r->ifname);

		list_del(&r->list);
		free(r);
	}
}

static void ev_handler(struct ubus_context *ctx,
					   struct ubus_event_handler *ev,
					   const char *type,
					   struct blob_attr *msg)
{
	struct json_object *json_msg;
	char ifname[IFNAMSIZ] = {0};
	char *str;
	const char *channel, *action;
	struct radio *r, *tmp;

	str = blobmsg_format_json(msg, true);
	if (!str)
		return;

	json_msg = json_tokener_parse(str);
	if (!json_msg)
		goto out_str;

	channel = json_get_string(json_msg, "radio");
	if (!channel)
		goto out_json;

	action = json_get_string(json_msg, "action");
	if (!action)
		goto out_json;

	/* If up event is seen we may stop flashing */
	if (strncmp(action, "up", IFNAMSIZ) != 0)
		goto out_json;

	list_for_each_entry_safe(r, tmp, &radios, list) {
		if (strncmp(r->ifname, channel, IFNAMSIZ) != 0)
			continue;

		DEBUG(LOG_DEBUG, "Interface %s returned, remove from radio list",
				r->ifname);

		list_del(&r->list);
		free(r);
	}

	if (list_empty(&radios))
		uloop_timeout_set(&flash_timeout, 0);

out_json:
	json_object_put(json_msg);
out_str:
	free(str);
}

static void reload_wireless_services(void)
{
	int obj, rv;
	struct blob_buf buf;

	if (!ctx) {
		ERROR("No ubus context allocated");
		goto out;
	}

	ubus_lookup_id(ctx, "uci", &obj);
	if (!obj) {
		ERROR("Couldn't lookup the id of ubus uci object");
		goto out;
	}

	memset(&buf, 0, sizeof(buf));
	rv = blob_buf_init(&buf, 0);
	if (rv) {
		ERROR("Couldn't initialize blob_buf");
		goto out;
	}

	rv = blobmsg_add_string(&buf, "config", "wireless");
	if (rv) {
		ERROR("Couldn't add string to blob_buf");
		goto out_buf;
	}

	rv = ubus_invoke(ctx, obj, "commit", buf.head, NULL, NULL, 5000);
	if (rv) {
		ERROR("Couldn't run ubus call uci commit %s",
				blobmsg_format_json(buf.head, true));
		goto out_buf;
	}

	rv = config_get_radios(&radios);
	if (rv)
		goto out_buf;

	rv = set_led_state("allflash");
	if (rv)
		goto out_buf;

	/* Stop flashing after 30 seconds if no event is recorded */
	uloop_timeout_set(&flash_timeout, 30 * 1000);

	ev.cb = ev_handler;
	ubus_register_event_handler(ctx, &ev, "wifi.radio");
	ubus_register_event_handler(ctx, &ev, "wifi.dfs");

out_buf:
	blob_buf_free(&buf);
out:
	return;
}

int remove_newline(char *str)
{
	char *pos;

	pos = strchr(str, '\n');
	if (pos) {
		*pos = '\0';
		return 0;
	}

	return -1;
}

/* will allocate memory! remember to free! */
char *get_file(char *path)
{
	struct blob_buf bb = {0};
	FILE *f;
	size_t len, nread;
	char *buffer;

	f = fopen(path, "r");
	if (!f)
		goto out;

	if (fseek(f, 0, SEEK_END))
		goto out_close;

	len = ftell(f);
	if (len < 0)
		goto out_close;

	if (fseek(f, 0, SEEK_SET))
		goto out_close;

	buffer = calloc(1, len);
	if (!buffer)
		goto out_close;

	nread = fread(buffer, sizeof(char), len, f);
	if (nread < len)
		goto out_free;

	fclose(f);
	return buffer;
out_free:
	free(buffer);
out_close:
	fclose(f);
out:
	return NULL;
}

static int netmode_status(struct ubus_context *ctx, struct ubus_object *obj,
		struct ubus_request_data *req, const char *method,
		struct blob_attr *msg)
{
	struct blob_buf bb = {0};
	FILE *f;
	size_t len;
	char *uplink_type, *curr_mode;

	curr_mode = get_file("/tmp/netmodes/old_mode");
	if (!curr_mode)
		goto out;

	blob_buf_init(&bb, 0);

	/* only interested in first line */
	remove_newline(curr_mode);

	blobmsg_add_string(&bb, "curr_mode", curr_mode);

	uplink_type = get_file("/tmp/netmodes/uplink_type");
	if (!uplink_type)
		goto out_print;

	/* only interested in first line */
	remove_newline(uplink_type);
	blobmsg_add_string(&bb, "uplink_type", uplink_type);

	free(uplink_type);
out_print:
	free(curr_mode);
	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);
out:
	return 0;
}

static int netmode_repeat(struct ubus_context *ctx, struct ubus_object *obj,
		struct ubus_request_data *req, const char *method,
		struct blob_attr *msg)
{
	int rv;
	enum config_compare_status compare_status;
	enum config_apply_status apply_status;
	struct blob_attr *tb[ARRAY_SIZE(netmode_repeat_policy)];
	char *wireless_cfg;
	struct blob_buf buf;
	const char *local_config_str, *remote_config_str, *network, *connect_bssid = NULL;
	struct json_object *local_config = NULL, *remote_config = NULL;
	struct json_object *remote_ifaces = NULL;

	DEBUG(LOG_DEBUG, "msg = \"%s\"", blobmsg_format_json(msg, true));

	rv = blobmsg_parse(netmode_repeat_policy,
			ARRAY_SIZE(netmode_repeat_policy),
			tb, blobmsg_data(msg), blobmsg_len(msg));
	if (rv) {
		ERROR("blobmsg_parse");
		return UBUS_STATUS_INVALID_ARGUMENT;
	}

	if (!tb[P_NETWORK] || !tb[P_WIRELESS])
		return UBUS_STATUS_INVALID_ARGUMENT;

	if (tb[P_CONNECT_BSSID])
		connect_bssid = blobmsg_data(tb[P_CONNECT_BSSID]);

	/* get the current local config */
	memset(&buf, 0, sizeof(buf));
	blob_buf_init(&buf, 0);
	rv = config_get_wireless(&buf);
	if (rv)
		goto fail;

	/* transform blob_bufs into json_objects */
	local_config_str = blobmsg_format_json(buf.head, true);
	remote_config_str = blobmsg_format_json(tb[P_WIRELESS], true);

	local_config = json_tokener_parse(local_config_str);
	remote_config = json_tokener_parse(remote_config_str);

	remote_ifaces = json_object_new_array();

	network = blobmsg_get_string(tb[P_NETWORK]);
	compare_status = config_compare_wireless(local_config, remote_config,
		remote_ifaces, network, connect_bssid);
	if (compare_status == CONFIG_COMPARE_EQUAL)
		goto out;
	if (compare_status == CONFIG_COMPARE_FAIL) {
		ERROR("config_compare_wireless failed");
		goto fail;
	}

	/* apply the new config then reload */
	apply_status = config_apply_wireless(remote_ifaces, connect_bssid, compare_status);
	if (apply_status == CONFIG_APPLY_FAIL) {
		ERROR("config_apply_wireless failed");
		goto fail;
	}

	reload_wireless_services();

out:
	json_object_put(local_config);
	json_object_put(remote_config);
	json_object_put(remote_ifaces);
	blob_buf_free(&buf);
	return UBUS_STATUS_OK;
fail:
	json_object_put(local_config);
	json_object_put(remote_config);
	json_object_put(remote_ifaces);
	blob_buf_free(&buf);
	return UBUS_STATUS_UNKNOWN_ERROR;
}

static int netmode_repeat_wifilife(struct ubus_context *ctx, struct ubus_object *obj,
						struct ubus_request_data *req, const char *method,
						struct blob_attr *msg)
{
	int rv;
	enum config_apply_status apply_status;
	struct blob_attr *tb[ARRAY_SIZE(netmode_repeat_policy)];
	struct blob_buf buf;
	const char *remote_config_str;
	struct json_object *remote_config = NULL, *remote_values;
	json_bool jrv;

	DEBUG(LOG_DEBUG, "msg = \"%s\"", blobmsg_format_json(msg, true));

	rv = blobmsg_parse(netmode_repeat_wifilife_policy,
			ARRAY_SIZE(netmode_repeat_wifilife_policy),
			tb, blobmsg_data(msg), blobmsg_len(msg));
	if (rv) {
		ERROR("blobmsg_parse");
		return UBUS_STATUS_INVALID_ARGUMENT;
	}

	if (!tb[P_WIFILIFE])
		return UBUS_STATUS_INVALID_ARGUMENT;

	/* get the current local config */
	memset(&buf, 0, sizeof(buf));
	blob_buf_init(&buf, 0);
	rv = config_get_wifilife(&buf);
	if (rv)
		goto fail;

	/* transform blob_bufs into json_objects */
	remote_config_str = blobmsg_format_json(tb[P_WIFILIFE], true);

	remote_config = json_tokener_parse(remote_config_str);

	DEBUG(LOG_DEBUG, "remote_wifilife:\n%s\n", json_object_to_json_string_ext(remote_config, JSON_C_TO_STRING_PRETTY));

	jrv = json_object_object_get_ex(remote_config, "values", &remote_values);
	if (!jrv) {
		ERROR("no values in remote_config");
		goto out;
	}

	apply_status = config_apply_wifilife(remote_values);
	if (apply_status == CONFIG_APPLY_FAIL) {
		ERROR("config_apply_wifilife ailed");
		goto fail;
	}

out:
	json_object_put(remote_config);
	blob_buf_free(&buf);
	return UBUS_STATUS_OK;
fail:
	blob_buf_free(&buf);
	return UBUS_STATUS_UNKNOWN_ERROR;
}

static int netmode_sync(struct ubus_context *ctx, struct ubus_object *obj,
		struct ubus_request_data *req, const char *method,
		struct blob_attr *msg)
{
	int rv = UBUS_STATUS_OK;

	if (pthread_mutex_trylock(&sync_mutex)) {
		perror("pthread_mutex_trylock");
		if (errno == EBUSY)
			rv = UBUS_STATUS_PERMISSION_DENIED;
		else
			rv = UBUS_STATUS_UNKNOWN_ERROR;
		goto out;
	}

	start_initial_timer(&sync_timer);

out:
	return rv;
}

static int netmode_sync_wifilife(struct ubus_context *ctx, struct ubus_object *obj,
		struct ubus_request_data *req, const char *method,
		struct blob_attr *msg)
{
	int rv = UBUS_STATUS_OK;

	if (pthread_mutex_trylock(&sync_mutex)) {
		perror("pthread_mutex_trylock");
		if (errno == EBUSY)
			rv = UBUS_STATUS_PERMISSION_DENIED;
		else
			rv = UBUS_STATUS_UNKNOWN_ERROR;
		goto out;
	}

	start_initial_timer_wifilife(&sync_timer);

out:
	return rv;
}

const char *short_options = "hi:v:q";

struct option long_options[] = {
	{"help", no_argument, NULL, 'h'},
	{"ignore", required_argument, NULL, 'i'},
	{"verbose", required_argument, NULL, 'v'},
	{"quiet", no_argument, NULL, 'q'}
};

static void usage(char *binary)
{
	if (LOG_LEVEL == LOG_NONE)
		return;
	fprintf(stderr,
		"Usage: %s [ <options> ... ]\n\n"
		"  -h | --help               Print usage\n"
		"  -i | --ignore <option>    Option in Wifi interfaces to "
							"ignore\n"
		"  -v | --verbose <level>    Set the verbosity level\n"
		"                     [debug, notice, warning, error]\n"
		"  -q | --quiet              suppress all output\n"
		"                     do not use with --verbose\n"
		"\n",
		binary);

	exit(0);
}

static int set_log_level(const char *level)
{
	if (strcmp(level, "debug") == 0)
		LOG_LEVEL = LOG_DEBUG;
	else if(strcmp(level, "notice") == 0)
		LOG_LEVEL = LOG_NOTICE;
	else if(strcmp(level, "warning") == 0)
		LOG_LEVEL = LOG_WARNING;
	else if(strcmp(level, "error") == 0)
		LOG_LEVEL = LOG_ERROR;
	else if(strcmp(level, "none") == 0)
		LOG_LEVEL = LOG_NONE;
	else
		return -1;
	DEBUG(LOG_DEBUG, "setting debug_level to %s %d", level, LOG_LEVEL);
	return 0;
}

static void parse_args(int argc, char **argv)
{
	int c;

	while (1) {
		c = getopt_long(argc, argv, short_options, long_options, NULL);
		if (c == -1)
			break;

		switch (c) {
		case 'h':
			usage(argv[0]);
			break;
		case 'i':
			config_ignore_option(optarg);
			break;
		case 'v':
			if (!optarg) {
				ERROR("missing verbosity level\n");
				usage(argv[0]);
				exit(99);
				break;
			}
			if (set_log_level(optarg) != 0) {
				ERROR("invalid verbosity level %s\n", optarg);
				usage(argv[0]);
				exit(99);
			}
			break;
		case 'q':
			set_log_level("none");
			break;
		default:
			usage(argv[0]);
			exit(99);
		}

	}
}

int main(int argc, char **argv)
{
	struct ubus_event_handler handler = {};
	struct uloop_timeout initial_wifilife_timer = {};
	struct uloop_timeout initial_timer = {};

	int rv;

	parse_args(argc, argv);

	INIT_LIST_HEAD(&radios);
	flash_timeout.cb = stop_flash_cb;

	rv = ubus_init();
	if (rv < 0) {
		ERROR("Failed to connect to ubus");
		return 1;
	}

	rv = ubus_add_object(ctx, &netmode_object);
	if (rv)
		ERROR("Failed to add object \"%s\" on ubus",
				netmode_object.name);

	rv = create_listener(&handler);
	if (rv)
		ERROR("Failed to initialize ubus listener");

	start_initial_timer(&initial_timer);
	start_initial_timer_wifilife(&initial_wifilife_timer);

	uloop_run();

	/* ubus_unregister_event_handler(ctx, &handler) */
	config_free_ignored_options();
	ubus_free(ctx);

	return 0;
}
